/* =============================================================================
 * conjunto2.h
 *
 * Estructura de datos y primitivas del TAD Conjunto - Versin 2.
 *
 * En esta versin los elementos de un conjunto son Elemento *.
 * Las primitivas del TAD son independientes del tipo de dato que almacenen los
 * elementos.
 * En esta implementacin del TAD, un elemento almacena un int.
 * Si se quiere tener un conjunto con elementos que no sean int, las primitivas
 * del TAD no hay que cambiarlas, pero habra que modificar la estructura de
 * datos y funciones de Elemento (ver elemento.h y elemento.c).
 *
 * Autor: Ivn Cantador
 * Fecha: 9/2/2017
 * Versin: 2.0
 * Comentarios: En este fichero habra que incluir las cabeceras de funcin,
 * describiendo las primitivas, sus argumentos de entrada y valores de salida.
 * ========================================================================== */

#include <stdio.h>
#include <stdlib.h>
#include "conjunto2.h"
#include "elemento.h"

/* -------------------------- Estructura de datos -------------------------- */
/* ------------------------------------------------------------------------- */

#define MAX_ELEMENTOS 1000

struct _Conjunto2 {
	/* Un elemento es un Elemento *, donde Elemento encapsula un dato de un tipo dado */
	Elemento *elementos[MAX_ELEMENTOS];
	int numElementos;
};

/* -------------------------- Funciones primitivas -------------------------- */
/* -------------------------------------------------------------------------- */

Conjunto2 *conjunto2_crear() {
	Conjunto2 *cjto = NULL;

	cjto = (Conjunto2 *) malloc(sizeof(Conjunto2));
	if ( cjto == NULL ) {
		return NULL;
	}

	cjto->numElementos = 0;

	return cjto;
}

void conjunto2_vaciar(Conjunto2 *cjto) {
	int i;

	if ( cjto != NULL ) {
		for ( i = 0 ; i < cjto->numElementos ; i++ ) {
			elemento_eliminar(cjto->elementos[i]);
		}
		cjto->numElementos = 0;
	}
}

void conjunto2_eliminar(Conjunto2 *cjto) {
	if ( cjto != NULL ) {
		conjunto2_vaciar(cjto);
		free(cjto);
	}
}

int conjunto2_cardinalidad(Conjunto2 *cjto) {
	if ( cjto == NULL ) {
		return -1;
	}

	return cjto->numElementos;
}

boolean conjunto2_estaVacio(Conjunto2 *cjto) {
	if ( cjto == NULL ) {
		return NULL_BOOLEAN;
	}

	return cjto->numElementos == 0 ? TRUE : FALSE;
}

boolean conjunto2_estaLleno(Conjunto2 *cjto) {
	if ( cjto == NULL ) {
		return NULL_BOOLEAN;
	}

	return cjto->numElementos == MAX_ELEMENTOS ? TRUE : FALSE;
}

Elemento *conjunto2_obtenerElemento(Conjunto2 *cjto, int pos) {
	Elemento *elem = NULL;

	if ( cjto == NULL || conjunto2_estaVacio(cjto) == TRUE || pos < 0 || pos >= cjto->numElementos ) {
		return NULL;
	}

	/* Devolvemos una copia del elemento en la posicin pos */
	/* La copia hay que liberarla fuera de esta funcion */
	elem = elemento_copiar(cjto->elementos[pos]);

	return elem;
}

int conjunto2_buscarElemento(Conjunto2 *cjto, Elemento *elem) {
	int i;

	if ( cjto == NULL || elem == NULL ) {
		return NAN;
	}

	/* Comparamos cada elemento del conjunto con el de entrada */
	for ( i = 0; i < cjto->numElementos; i++ ) {
		if ( elemento_comparar(cjto->elementos[i], elem) == 0 ) {
			return i;
		}
	}

	return -1;
}

status conjunto2_insertarElemento(Conjunto2 *cjto, Elemento *elem) {
	if ( cjto == NULL || conjunto2_estaLleno(cjto) == TRUE || elem == NULL ) {
		return ERROR;
	}

	if ( conjunto2_buscarElemento(cjto, elem) != -1 ) {
		return OK;
	}

	/* Almacenamos una copia del elemento de entrada */
	cjto->elementos[cjto->numElementos] = elemento_copiar(elem);
	if ( cjto->elementos[cjto->numElementos] == NULL ) {
		return ERROR;
	}

	cjto->numElementos++;

	return OK;
}

void conjunto2_visualizar(Conjunto2 *cjto) {
	void *e = NULL;
	int i, n;

	if ( cjto != NULL ) {
		n = conjunto2_cardinalidad(cjto);
		for ( i=0; i<n; i++ ) {
			e = conjunto2_obtenerElemento(cjto, i);
			elemento_visualizar(e);
			elemento_eliminar(e);
		}
		printf("\n");
	}
}

Conjunto2 *conjunto2_union(Conjunto2 *cjtoA, Conjunto2 *cjtoB) {
	Conjunto2 *cjtoC = NULL;
	void *e = NULL;
	int i, n;

	if ( cjtoA == NULL || cjtoB == NULL ) {
		return NULL;
	}

	cjtoC = conjunto2_crear();
	if ( cjtoC == NULL ) {
		return NULL;
	}

	n = conjunto2_cardinalidad(cjtoA);
	for ( i=0; i<n; i++ ) {
		e = conjunto2_obtenerElemento(cjtoA, i);
		conjunto2_insertarElemento(cjtoC, e);
		elemento_eliminar(e);
	}

	n = conjunto2_cardinalidad(cjtoB);
	for ( i=0; i<n; i++ ) {
		e = conjunto2_obtenerElemento(cjtoB, i);
		if ( conjunto2_buscarElemento(cjtoA, e) == -1 ) {
			conjunto2_insertarElemento(cjtoC, e);
		}
		elemento_eliminar(e);
	}

	return cjtoC;
}

Conjunto2 *conjunto2_interseccion(Conjunto2 *cjtoA, Conjunto2 *cjtoB) {
	Conjunto2 *cjtoC = NULL;
	void *e = NULL;
	int i, n;

	if ( cjtoA == NULL || cjtoB == NULL ) {
		return NULL;
	}

	cjtoC = conjunto2_crear();
	if ( cjtoC == NULL ) {
		return NULL;
	}

	n = conjunto2_cardinalidad(cjtoA);
	for ( i=0; i<n; i++ ) {
		e = conjunto2_obtenerElemento(cjtoA, i);
		if ( conjunto2_buscarElemento(cjtoB, e) != -1 ) {
			conjunto2_insertarElemento(cjtoC, e);
		}
		elemento_eliminar(e);
	}

	return cjtoC;
}
