/* =============================================================================
 * conjunto3.c
 *
 * Estructura de datos y primitivas del TAD Conjunto - Versin 3.
 *
 * En esta versin los elementos de un conjunto son void *.
 * Las primitivas del TAD son independientes del tipo de dato que almacenen los
 * elementos.
 * En esta implementacin del TAD, los datos de un conjunto son opacos y se
 * manejan por una serie de funciones que se almacenan en la estrucutra Conjunto.
 * Estas funciones se pasan como argumentos de entrada en conjunto3_crear.
 *
 * Autor: Ivn Cantador
 * Fecha: 9/2/2017
 * Versin: 3.0
 * Comentarios: En este fichero habra que incluir las cabeceras de funcin,
 * describiendo las primitivas, sus argumentos de entrada y valores de salida.
 * ========================================================================== */

#include <stdio.h>
#include <stdlib.h>
#include "conjunto3.h"

/* -------------------------- Estructura de datos -------------------------- */
/* ------------------------------------------------------------------------- */

#define MAX_ELEMENTOS 1000

struct _Conjunto3 {
	/* Un elemento es un void *, que podr a apuntar a cualquier tipo de dato */
	void *elementos[MAX_ELEMENTOS];
	int numElementos;

	/* Punteros a funcion para copiar, comparar, eliminar y visualizar por pantalla elementos */
	void *(*funcion_copiarElemento)(void *);
	int (*funcion_compararElementos)(void *, void *);
	void (*funcion_eliminarElemento)(void *);
	void (*funcion_visualizarElemento)(void *);
};

/* -------------------------- Funciones primitivas -------------------------- */
/* -------------------------------------------------------------------------- */

Conjunto3 *conjunto3_crear(void *(*fnc_copiar)(void *), int (*fnc_comparar)(void *, void *), void (*fnc_eliminar)(void *), void (*fnc_visualizar)(void *)) {
	Conjunto3 *cjto = NULL;

	if ( fnc_copiar == NULL || fnc_comparar == NULL || fnc_eliminar == NULL || fnc_visualizar == NULL ) {
		return NULL;
	}

	cjto = (Conjunto3 *) malloc(sizeof(Conjunto3));
	if ( cjto == NULL ) {
		return NULL;
	}

	cjto->numElementos = 0;

	cjto->funcion_copiarElemento =  fnc_copiar;
	cjto->funcion_compararElementos = fnc_comparar;
	cjto->funcion_eliminarElemento =  fnc_eliminar;
	cjto->funcion_visualizarElemento = fnc_visualizar;

	return cjto;
}

void conjunto3_vaciar(Conjunto3 *cjto) {
	int i;

	if ( cjto != NULL ) {
		for ( i = 0 ; i < cjto->numElementos ; i++ ) {
			cjto->funcion_eliminarElemento(cjto->elementos[i]);
		}
		cjto->numElementos = 0;
	}
}

void conjunto3_eliminar(Conjunto3 *cjto) {
	if ( cjto != NULL ) {
		conjunto3_vaciar(cjto);
		free(cjto);
	}
}

int conjunto3_cardinalidad(Conjunto3 *cjto) {
	if ( cjto == NULL ) {
		return -1;
	}

	return cjto->numElementos;
}

boolean conjunto3_estaVacio(Conjunto3 *cjto) {
	if ( cjto == NULL ) {
		return NULL_BOOLEAN;
	}

	return cjto->numElementos == 0 ? TRUE : FALSE;
}

boolean conjunto3_estaLleno(Conjunto3 *cjto) {
	if ( cjto == NULL ) {
		return NULL_BOOLEAN;
	}

	return cjto->numElementos == MAX_ELEMENTOS ? TRUE : FALSE;
}

void *conjunto3_obtenerElemento(Conjunto3 *cjto, int pos) {
	void *elem = NULL;

	if ( cjto == NULL || conjunto3_estaVacio(cjto) == TRUE || pos < 0 || pos >= cjto->numElementos ) {
		return NULL;
	}

	/* Devolvemos una copia del elemento en la posicin pos */
	/* La copia hay que liberarla fuera de esta funcion */
	elem = cjto->funcion_copiarElemento(cjto->elementos[pos]);

	return elem;
}

int conjunto3_buscarElemento(Conjunto3 *cjto, void *elem) {
	int i;

	if ( cjto == NULL || elem == NULL ) {
		return NAN;
	}

	/* Comparamos cada elemento del conjunto con el de entrada */
	for ( i = 0; i < cjto->numElementos; i++ ) {
		if ( cjto->funcion_compararElementos(cjto->elementos[i], elem) == 0 ) {
			return i;
		}
	}

	return -1;
}

status conjunto3_insertarElemento(Conjunto3 *cjto, void *elem) {
	if ( cjto == NULL || conjunto3_estaLleno(cjto) == TRUE || elem == NULL ) {
		return ERROR;
	}

	if ( conjunto3_buscarElemento(cjto, elem) != -1 ) {
		return OK;
	}

	/* Almacenamos una copia del elemento de entrada */
	cjto->elementos[cjto->numElementos] = cjto->funcion_copiarElemento(elem);
	if ( cjto->elementos[cjto->numElementos] == NULL ) {
		return ERROR;
	}

	cjto->numElementos++;

	return OK;
}

void conjunto3_visualizar(Conjunto3 *cjto) {
	void *e = NULL;
	int i, n;

	if ( cjto != NULL ) {
		n = conjunto3_cardinalidad(cjto);
		for ( i=0; i<n; i++ ) {
			e = conjunto3_obtenerElemento(cjto, i);
			cjto->funcion_visualizarElemento(e);
			cjto->funcion_eliminarElemento(e);
		}
		printf("\n");
	}
}

Conjunto3 *conjunto3_union(Conjunto3 *cjtoA, Conjunto3 *cjtoB) {
	Conjunto3 *cjtoC = NULL;
	void *e = NULL;
	int i, n;

	if ( cjtoA == NULL || cjtoB == NULL ) {
		return NULL;
	}

	cjtoC = conjunto3_crear(cjtoA->funcion_copiarElemento, cjtoA->funcion_compararElementos, cjtoA->funcion_eliminarElemento, cjtoA->funcion_visualizarElemento);
	if ( cjtoC == NULL ) {
		return NULL;
	}

	n = conjunto3_cardinalidad(cjtoA);
	for ( i=0; i<n; i++ ) {
		e = conjunto3_obtenerElemento(cjtoA, i);
		conjunto3_insertarElemento(cjtoC, e);
		cjtoC->funcion_eliminarElemento(e);
	}

	n = conjunto3_cardinalidad(cjtoB);
	for ( i=0; i<n; i++ ) {
		e = conjunto3_obtenerElemento(cjtoB, i);
		if ( conjunto3_buscarElemento(cjtoA, e) == -1 ) {
			conjunto3_insertarElemento(cjtoC, e);
		}
		cjtoC->funcion_eliminarElemento(e);
	}

	return cjtoC;
}

Conjunto3 *conjunto3_interseccion(Conjunto3 *cjtoA, Conjunto3 *cjtoB) {
	Conjunto3 *cjtoC = NULL;
	void *e = NULL;
	int i, n;

	if ( cjtoA == NULL || cjtoB == NULL ) {
		return NULL;
	}

	cjtoC = conjunto3_crear(cjtoA->funcion_copiarElemento, cjtoA->funcion_compararElementos, cjtoA->funcion_eliminarElemento, cjtoA->funcion_visualizarElemento);
	if ( cjtoC == NULL ) {
		return NULL;
	}

	n = conjunto3_cardinalidad(cjtoA);
	for ( i=0; i<n; i++ ) {
		e = conjunto3_obtenerElemento(cjtoA, i);
		if ( conjunto3_buscarElemento(cjtoB, e) != -1 ) {
			conjunto3_insertarElemento(cjtoC, e);
		}
		cjtoC->funcion_eliminarElemento(e);
	}

	return cjtoC;
}



